"""
The ``solvers`` module provides interfaces for solving the systems of ordinary
differential equations that govern equivalent circuit models. The solvers are
taken from `scikit-sundae`_, a wrapper for `SUNDIALS`_. Specifically, both the
IDA (for differential-algebraic equations) and CVODE (for ordinary differential
equations) interfaces are included.

These solvers offer robust numerical methods with settings to adjust the
accuracy, stability, and performance. While the implementations from from the
``scikit-sundae`` package, they are included here to centralize documentation
and streamline solver configuration within the ECM framework.

.. _scikit-sundae: https://scikit-sundae.readthedocs.io
.. _SUNDIALS: https://sundials.readthedocs.io

"""

from ._cvode import CVODEResult, CVODESolver
from ._ida import IDAResult, IDASolver

__all__ = [
    'CVODEResult',
    'CVODESolver',
    'IDAResult',
    'IDASolver',
]



from __future__ import annotations
from typing import Iterable, Callable, TypeVar, TYPE_CHECKING

import textwrap
from copy import deepcopy

import atexit
import numpy as np
import matplotlib.pyplot as plt

from .solvers import IDAResult

if TYPE_CHECKING:  # pragma: no cover
    from ._simulation import Simulation

    Solution = TypeVar('Solution', bound='BaseSolution')

if not hasattr(np, 'concat'):  # pragma: no cover
    np.concat = np.concatenate


class ExitHandler:
    """
    Exit handler.

    Use this class to register functions that you want to run just before a
    file exits. This is primarily used to register plt.show() so plots appear
    in both interactive and non-interactive environments, even if the user
    forgets to explicitly call it.

    """
    _registered = []

    @classmethod
    def register_atexit(cls, func: Callable) -> None:
        if func not in cls._registered:
            atexit.register(func)


class BaseSolution(IDAResult):
    """Base solution."""

    def __init__(self) -> None:
        """
        The base solution class is a parent class to both the StepSolution
        and CycleSolution classes. Inheriting from this class gives each
        solution instance a 'vars' dictionary, access to the 'plot' method,
        and ensures that the slicing of the solution vector into 'vars' is
        consistent between all solutions.

        """

        self.vars = {}

    def __repr__(self) -> str:  # pragma: no cover
        """
        Return a readable repr string.

        Returns
        -------
        readable : str
            A console-readable instance representation.

        """

        classname = self.__class__.__name__

        def wrap_string(label: str, value: list, width: int):
            if isinstance(value, Iterable):
                value = list(value)
            else:
                value = [value]

            indent = ' '*(len(label) + 1)

            if classname == 'StepSolution' and len(value) == 1:
                text = label + f"{value[0]!r}"
            else:
                text = label + "[" + ", ".join(f"{v!r}" for v in value) + "]"

            return textwrap.fill(text, width=width, subsequent_indent=indent)

        data = [
            wrap_string('    success=', self.success, 79),
            wrap_string('    status=', self.status, 79),
            wrap_string('    nfev=', self.nfev, 79),
            wrap_string('    njev=', self.njev, 79),
            wrap_string('    vars=', self.vars.keys(), 79),
        ]

        summary = f"    solvetime={self.solvetime},"
        for d in data:
            summary += f"\n{d},"

        readable = f"{classname}(\n{summary}\n)"

        return readable

    def plot(self, x: str, y: str, **kwargs) -> None:
        """
        Plot any two variables in 'vars' against each other.

        Parameters
        ----------
        x : str
            A variable key in 'vars' to be used for the x-axis.
        y : str
            A variable key in 'vars' to be used for the y-axis.
        **kwargs : dict, optional
            Keyword arguments to pass through to `plt.plot()`. For more info
            please refer to documentation for `maplotlib.pyplot.plot()`.

        Returns
        -------
        None.

        """

        plt.figure()
        plt.plot(self.vars[x], self.vars[y], **kwargs)

        if '_' in x:
            variable, units = x.split('_')
            xlabel = variable.capitalize() + ' [' + units + ']'
        else:
            xlabel = x

        if '_' in y:
            variable, units = y.split('_')
            ylabel = variable.capitalize() + ' [' + units + ']'
        else:
            ylabel = y

        plt.xlabel(xlabel)
        plt.ylabel(ylabel)

        if not plt.isinteractive():
            ExitHandler.register_atexit(plt.show)

    def _fill_vars(self) -> None:
        """
        Fills the 'vars' dictionary by slicing the SolverReturn solution
        states. Users should generally only access the solution via 'vars'
        since names are more intuitive than interpreting 'y' directly.

        Returns
        -------
        None.

        """

        from ._basemodel import calculated_current

        sim = self._sim
        ptr = sim._ptr

        time = self.t

        soc = self.y[:, ptr['soc']]
        T_cell = self.y[:, ptr['T_cell']]*sim._T_ref
        hyst = self.y[:, ptr['hyst']]
        eta_j = self.y[:, ptr['eta_j']]
        voltage = self.y[:, ptr['V_cell']]

        try:
            ocv = sim.ocv(soc)
            R0 = sim.R0(soc, T_cell)

            assert isinstance(ocv, np.ndarray)
            assert isinstance(R0, np.ndarray)
            assert ocv.shape == soc.shape
            assert R0.shape == soc.shape

        except (TypeError, AssertionError):

            ocv = np.empty_like(soc)
            R0 = np.empty_like(soc)
            for i in range(soc.size):
                ocv[i] = sim.ocv(soc[i])
                R0[i] = sim.R0(soc[i], T_cell[i])

        current = calculated_current(voltage, ocv, hyst, eta_j, R0)

        # stored time
        self.vars['time_s'] = time
        self.vars['time_min'] = time / 60.
        self.vars['time_h'] = time / 3600.

        # from state variables
        self.vars['soc'] = soc
        self.vars['temperature_K'] = T_cell
        self.vars['voltage_V'] = voltage
        self.vars['hysteresis_V'] = hyst

        # post-processed variables
        self.vars['current_A'] = current
        self.vars['power_W'] = current*voltage
        self.vars['eta0_V'] = current*R0

        for j, eta in enumerate(eta_j.T, start=1):
            self.vars['eta' + str(j) + '_V'] = eta


class StepSolution(BaseSolution):
    """Single-step solution."""

    def __init__(self, sim: Simulation, ida_soln: IDAResult,
                 timer: float) -> None:
        """
        A solution instance for a single experimental step.

        Parameters
        ----------
        sim : Simulation
            The simulation instance that was run to produce the solution.
        ida_soln : IDAResult
            The unformatted solution returned by IDASolver.
        timer : float
            Amount of time it took for IDASolver to perform the integration.

        """

        super().__init__()

        self._sim = deepcopy(sim)

        self.message = ida_soln.message
        self.success = ida_soln.success
        self.status = ida_soln.status

        self.t = ida_soln.t
        self.y = ida_soln.y
        self.yp = ida_soln.yp

        self.i_events = ida_soln.i_events
        self.t_events = ida_soln.t_events
        self.y_events = ida_soln.y_events
        self.yp_events = ida_soln.yp_events

        self.nfev = ida_soln.nfev
        self.njev = ida_soln.njev

        self._timer = timer

        self._fill_vars()

    @property
    def solvetime(self) -> str:
        """
        Print a statement specifying how long IDASolver spent integrating.

        Returns
        -------
        solvetime : str
            An f-string with the solver integration time in seconds.

        """
        return f"{self._timer:.3f} s"


class CycleSolution(BaseSolution):
    """All-step solution."""

    def __init__(self, *soln: StepSolution, t_shift: float = 1e-3) -> None:
        """
        A solution instance with all experiment steps stitch together into
        a single cycle.

        Parameters
        ----------
        *soln : StepSolution
            All unpacked StepSolution instances to stitch together. The given
            steps should be given in the same sequential order that they were
            run.
        t_shift : float
            Time (in seconds) to shift step solutions by when stitching them
            together. If zero the end time of each step overlaps the starting
            time of its following step. The default is 1e-3.

        """

        super().__init__()

        soln = deepcopy(list(soln))  # ensure type list, memory safe copy

        self._solns = soln
        self._sim = soln[0]._sim

        t_size = np.sum([soln.t.size for soln in self._solns])
        sv_size = self._sim._sv0.size

        self.message = []
        self.success = []
        self.status = []

        self.t = np.empty([t_size])
        self.y = np.empty([t_size, sv_size])
        self.yp = np.empty([t_size, sv_size])

        self.t_events = None
        self.y_events = None
        self.yp_events = None

        self.nfev = []
        self.njev = []

        self._timers = []

        first = 0
        for soln in self._solns:
            soln_size = soln.t.size
            last = first + soln_size

            if first > 0:
                shift_t = self.t[first - 1] + soln.t + t_shift
            else:
                shift_t = soln.t

            if soln.t_events is not None and first > 0:
                shift_t_events = self.t[first - 1] + soln.t_events + t_shift
            elif soln.t_events is not None:
                shift_t_events = soln.t_events

            self.message.append(soln.message)
            self.success.append(soln.success)
            self.status.append(soln.status)

            self.t[first:last] = shift_t
            self.y[first:last, :] = soln.y
            self.yp[first:last, :] = soln.yp

            first = last

            if soln.t_events is not None:
                if self.t_events is None:
                    self.t_events = shift_t_events
                    self.y_events = soln.y_events
                    self.yp_events = soln.yp_events
                else:
                    self.t_events = np.concat([self.t_events, shift_t_events])
                    self.y_events = np.concat([self.y_events, soln.y_events])
                    self.yp_events = np.concat([self.yp_events, soln.yp_events])

            self.nfev.append(soln.nfev)
            self.njev.append(soln.njev)

            self._timers.append(soln._timer)

        self._fill_vars()

    @property
    def solvetime(self) -> str:
        """
        Print a statement specifying how long IDASolver spent integrating.

        Returns
        -------
        solvetime : str
            An f-string with the total solver integration time in seconds.

        """
        return f"{sum(self._timers):.3f} s"

    def get_steps(self, idx: int | tuple) -> StepSolution | CycleSolution:
        """
        Return a subset of the solution.

        Parameters
        ----------
        idx : int | tuple
            The step index (int) or range (first, last) to return.

        Returns
        -------
        :class:`StepSolution` | :class:`CycleSolution`
            The returned solution subset. A StepSolution is returned if 'idx'
            is an int, and a CycleSolution will be returned for the range of
            requested steps when 'idx' is a tuple.

        """

        if isinstance(idx, int):
            return deepcopy(self._solns[idx])
        elif isinstance(idx, (tuple, list)):
            solns = self._solns[idx[0]:idx[1] + 1]
            return CycleSolution(*solns)

    def append_soln(self, soln: Solution, t_shift: float = 1e-3) -> None:
        """
        Append another solution object to the current instance. Appending the
        instance itself is also allowed, which is helpful to visualize multiple
        cycles.

        Parameters
        ----------
        soln : StepSolution | CycleSolution
            A solution (step or cycle) to append to the end of the current
            instance.
        t_shift : float, optional
            Time (in seconds) to add between the current solution and new
            appended solution. If zero, the current final time and new initial
            time will exactly overlap. The default is 1e-3.

        Returns
        -------
        None.

        Raises
        ------
        TypeError
            'soln' input must be StepSolution, CycleSolution.
        ValueError
            'soln' input is incompatible because it came from a simulation with
            different num_RC_pairs.

        Notes
        -----
        The ``t_shift`` input only affects the shift between the current and
        new solutions. For example if two ``CycleSolution`` had original shifts
        of 1 and 5 for their respective steps when stitched, this appended case
        will still have 1 for the first few steps, ``t_shift`` between, and 5
        for the shifts between the latter half's steps.

        Appending solutions cannot simply be undone. If you think you may want
        to go back to an instance prior to appending another solution then you
        should create a copy to operate on. Use ``copy.deepcopy`` from Python's
        standard library to make sure the copy is memory safe, i.e., the two
        instances do not share any common memory.

        """

        if not isinstance(soln, (StepSolution, CycleSolution)):
            raise TypeError("'soln' input must be StepSolution, CycleSolution.")

        if self._sim.num_RC_pairs != soln._sim.num_RC_pairs:
            raise ValueError("'soln' input is incompatible because it came"
                             " from a simulation with different num_RC_pairs.")

        soln = deepcopy(soln)

        shift_t = self.t[-1] + soln.t + t_shift
        if soln.t_events is not None:
            shift_t_events = self.t[-1] + soln.t_events + t_shift

        if isinstance(soln, StepSolution):
            self._solns.append(soln)

            self.message.append(soln.message)
            self.success.append(soln.success)
            self.status.append(soln.status)

            self.nfev.append(soln.nfev)
            self.njev.append(soln.njev)

            self._timers.append(soln._timer)

        elif isinstance(soln, CycleSolution):
            self._solns.extend(soln._solns)

            self.message.extend(soln.message)
            self.success.extend(soln.success)
            self.status.extend(soln.status)

            self.nfev.extend(soln.nfev)
            self.njev.extend(soln.njev)

            self._timers.extend(soln._timers)

        self.t = np.concat([self.t, shift_t])
        self.y = np.concat([self.y, soln.y])
        self.yp = np.concat([self.yp, soln.yp])

        if soln.t_events is not None:
            if self.t_events is None:
                self.t_events = shift_t_events
                self.y_events = soln.y_events
                self.yp_events = soln.yp_events
            else:
                self.t_events = np.concat([self.t_events, shift_t_events])
                self.y_events = np.concat([self.y_events, soln.y_events])
                self.yp_events = np.concat([self.yp_events, soln.yp_events])

        self._fill_vars()



from __future__ import annotations

import re
import os
import warnings
from abc import ABC, abstractmethod

import numpy as np
from ruamel.yaml import YAML, add_constructor, SafeConstructor


def calculated_current(voltage, ocv, hyst, eta_j, R0) -> float | np.ndarray:
    """
    Calculate the current [A].

    Inputs must be consistent with one another. For example, to calculate the
    current at a single time all values must be a float except eta_j which is
    a 1D array at a given time step. To return the current across all times,
    use 1D arrays for all inputs except eta_j, which then must be a 2D array
    with rows corresponding to times and columns to each RC pair. eta_j can
    be empty if there are no RC pairs, but must still be 1D or 2D as required.

    Parameters
    ----------
    voltage : float | np.ndarray
        Cell voltage at a single or multiple times [V].
    ocv : float | np.ndarray
        Open-circuit voltage at a single or multiple times [V].
    hyst : float | np.ndarray
        Hysteresis voltage at a single or multiple times [V].
    eta_j : np.ndarray
        RC-pair overpotentials at a single (1D) or multiple (2D) times [V]. If
        2D the rows correspond to time indices and columns to each RC pair.
    R0 : float | np.ndarray
        Series resistance values at a single or multiple times [Ohm].

    Returns
    -------
    current : float | np.ndarray
        Calculated current at a single or multiple times [A].

    """

    if eta_j.ndim == 1:
        return -(voltage - ocv - hyst + np.sum(eta_j)) / R0
    elif eta_j.ndim == 2:
        return -(voltage - ocv - hyst + np.sum(eta_j, axis=1)) / R0
    else:
        raise ValueError("Dimension error in calculating current.")


def calculated_voltage(current, ocv, hyst, eta_j, R0) -> float | np.ndarray:
    """
    Calculate the voltage [V].

    Inputs must be consistent with one another. For example, to calculate the
    voltage at a single time all values must be a float except eta_j which is
    a 1D array at a given time step. To return the voltage across all times,
    use 1D arrays for all inputs except eta_j, which then must be a 2D array
    with rows corresponding to times and columns to each RC pair. eta_j can
    be empty if there are no RC pairs, but must still be 1D or 2D as required.

    Parameters
    ----------
    current : float | np.ndarray
        Current at a single or multiple times [A].
    ocv : float | np.ndarray
        Open-circuit voltage at a single or multiple times [V].
    hyst : float | np.ndarray
        Hysteresis voltage at a single or multiple times [V].
    eta_j : np.ndarray
        RC-pair overpotentials at a single (1D) or multiple (2D) times [V]. If
        2D the rows correspond to time indices and columns to each RC pair.
    R0 : float | np.ndarray
        Series resistance values at a single or multiple times [Ohm].

    Returns
    -------
    voltage : float | np.ndarray
        Calculated voltage at a single or multiple times [V].

    """

    if eta_j.ndim == 1:
        return ocv + hyst - np.sum(eta_j) - current*R0
    elif eta_j.ndim == 2:
        return ocv + hyst - np.sum(eta_j, axis=1) - current*R0
    else:
        raise ValueError("Dimension error in calculating voltage.")


class BaseModel(ABC):
    """Abstract BaseModel for Simulation and Prediction."""

    def __init__(self, params: dict | str = 'params.yaml'):
        """
        Models can be constructed using either a dictionary or a '.yaml' file.
        Note that the number of Rj and Cj attributes must be consistent with
        the 'num_RC_pairs' value. See the notes for more information on the
        callable parameters.

        Parameters
        ----------
        params : dict | str
            Mapping of model parameter names to their values. Can be either
            a dict or absolute/relative file path to a yaml file (str). The
            keys/value pair descriptions are given below. The default uses an
            internal yaml file.

            ============= ========================== ================
            Key           Value                      *type*, units
            ============= ========================== ================
            num_RC_pairs  number of RC pairs         *int*, -
            soc0          initial state of charge    *float*, -
            capacity      maximum battery capacity   *float*, Ah
            ce            coulombic efficiency       *float*, -
            gamma         hysteresis approach rate   *float*, -
            mass          total battery mass         *float*, kg
            isothermal    flag for isothermal model  *bool*, -
            Cp            specific heat capacity     *float*, J/kg/K
            T_inf         room/air temperature       *float*, K
            h_therm       convective coefficient     *float*, W/m2/K
            A_therm       heat loss area             *float*, m2
            ocv           open circuit voltage       *callable*, V
            M_hyst        max hysteresis magnitude   *callable*, V
            R0            series resistance          *callable*, Ohm
            Rj            resistance in RCj          *callable*, Ohm
            Cj            capacity in RCj            *callable*, F
            ============= ========================== ================

        Raises
        ------
        TypeError
            'params' must be type dict or str.
        ValueError
            'params' contains invalid and/or excess key/value pairs.

        Notes
        -----
        The 'ocv' and 'M_hyst' properties need to be callables with signatures
        like ``f(soc: float) -> float``, where 'soc' is the state of charge.
        All other properties that require callables (e.g., R0, Rj, and Cj) need
        signatures like ``f(soc: float, T_cell: float) -> float``, with 'T_cell'
        being the cell temperature in K.

        Rj and Cj are not real property names. These are used generally in the
        documentation. If ``num_RC_pairs=1`` then in addition to R0, you should
        define R1 and C1. If ``num_RC_pairs=2`` then you should also give R2
        and C2, etc. For the special case where ``num_RC_pairs=0``, you should
        not provide any resistance or capacitance values besides the series
        resistance R0, which is always required.

        While most parameters can be dynamically updated, the ``num_RC_pairs``
        attribute is read-only. Consequently, you cannot add nor remove Rj and
        Cj attributes. However, modifying the values of Rj and Cj functions
        is allowed. If you need a circuit with a different number of RC pairs
        then you will need to create a separate instance.

        """

        if isinstance(params, dict):
            params = params.copy()
        elif isinstance(params, str):
            params = _yaml_reader(params)

        self._repr_keys = [
            'num_RC_pairs',
            'soc0',
            'capacity',
            'ce',
            'gamma',
            'mass',
            'isothermal',
            'Cp',
            'T_inf',
            'h_therm',
            'A_therm',
        ]

        self._num_RC_pairs = params.pop('num_RC_pairs')
        self.soc0 = params.pop('soc0')
        self.capacity = params.pop('capacity')
        self.ce = params.pop('ce')
        self.gamma = params.pop('gamma')
        self.mass = params.pop('mass')
        self.isothermal = params.pop('isothermal')
        self.Cp = params.pop('Cp')
        self.T_inf = params.pop('T_inf')
        self.h_therm = params.pop('h_therm')
        self.A_therm = params.pop('A_therm')

        self.ocv = params.pop('ocv')
        self.R0 = params.pop('R0')

        for j in range(1, self.num_RC_pairs + 1):

            assert 'R' + str(j) in params, f"'params' is missing R{str(j)}"
            assert 'C' + str(j) in params, f"'params' is missing C{str(j)}"

            setattr(self, 'R' + str(j), params.pop('R' + str(j)))
            setattr(self, 'C' + str(j), params.pop('C' + str(j)))

        self.M_hyst = params.pop('M_hyst')

        if len(params) != 0:
            extra_keys = list(params.keys())
            raise ValueError("'params' contains invalid and/or excess"
                             f" key/value pairs: {extra_keys=}")

        self._T_ref = self.T_inf

        self.pre()

    def __repr__(self) -> str:  # pragma: no cover
        """
        Return a readable repr string.

        Returns
        -------
        readable : str
            A console-readable instance representation.

        """

        keys = self._repr_keys
        values = [getattr(self, k) for k in keys]

        summary = "\n    ".join([f"{k}={v}," for k, v in zip(keys, values)])
        readable = f"{self.classname}(\n    {summary}\n)"

        return readable

    @property
    def classname(self) -> str:
        """Return the name of the class."""
        return self.__class__.__name__

    @property
    def num_RC_pairs(self) -> int:
        """Return the number of RC pairs."""
        return self._num_RC_pairs

    @property
    def _classname(self) -> str:
        """Use until 'Model' is deprecated to control flags in _rhsfn."""
        if self.classname == 'Model':
            return 'Simulation'

        return self.classname

    @abstractmethod
    def pre(self, *args, **kwargs) -> None:  # pragma: no cover
        """Preprocessor: ensure model setup is correct and ready to run."""
        pass

    def _check_RC_pairs(self) -> None:
        """Verify correct attributes are set based on num_RC_pairs."""

        missing_attrs = []
        for j in range(1, self.num_RC_pairs + 1):
            if not hasattr(self, 'R' + str(j)):
                missing_attrs.append('R' + str(j))
            if not hasattr(self, 'C' + str(j)):
                missing_attrs.append('C' + str(j))

        classname = self.classname
        if missing_attrs:
            raise AttributeError(f"'{classname}' missing attrs {missing_attrs}"
                                 " to be consistent with 'num_RC_pairs'.")

        extra_attrs = []
        pattern = re.compile(r"^[RC](\d+)")
        for attr in list(self.__dict__.keys()):

            matches = pattern.match(attr)
            if matches is None:
                pass
            elif int(matches.group(1)) > self.num_RC_pairs:
                extra_attrs.append(attr)

        if extra_attrs:
            short_warn(f"Extra RC attributes {extra_attrs} are present, beyond"
                       " what was expected based on 'num_RC_pairs'.")

    def _rhsfn(self, t: float, sv: np.ndarray, userdata: dict) -> np.ndarray:
        """
        Right-hand-side functions.

        Given a DAE of the form ``M*yp = f(t, y)``, this function represents
        the right-hand-side ``f(t, y)``. For the 'Simulation' class, M has a
        single zero-element diagonal term for the algebraic expression used to
        solve for the cell voltage, making the system a DAE. When the system
        of equations is accessed via the 'Prediction' class, only the pure ODE
        is returned, allowing for faster solves via CVODE.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        userdata : dict
            Dictionary detailing an experimental step.

        Returns
        -------
        rhs : 1D np.array
            The right-hand-side array: equivalent to yp for the 'Prediction'
            class, but contains an extra algebraic expression for 'Simulation'.

        """

        ptr = self._ptr
        rhs = np.zeros(ptr['size'])

        # state
        soc = sv[ptr['soc']]
        T_cell = sv[ptr['T_cell']]*self._T_ref
        hyst = sv[ptr['hyst']]
        eta_j = sv[ptr['eta_j']]

        # state-dependent properties
        ocv = self.ocv(soc)
        R0 = self.R0(soc, T_cell)

        # dependent parameters
        Q_inv = 1. / (3600. * self.capacity)
        alpha_inv = 1. / (self.mass * self.Cp * self._T_ref)

        # current, voltage, and power - different for Simulation/Prediction
        if self._classname == 'Simulation':
            voltage = sv[ptr['V_cell']]
            current = calculated_current(voltage, ocv, hyst, eta_j, R0)

        elif self._classname == 'Prediction':
            current = userdata['current'](t)
            voltage = calculated_voltage(current, ocv, hyst, eta_j, R0)

        power = current*voltage

        # state of charge (differential)
        ce = 1. if current >= 0. else self.ce
        rhs[ptr['soc']] = -ce*current*Q_inv

        # temperature (differential)
        Q_gen = current*(ocv + hyst - voltage)
        Q_conv = self.h_therm*self.A_therm*(self.T_inf - T_cell)

        rhs[ptr['T_cell']] = alpha_inv * (Q_gen + Q_conv) \
                           * (1 - self.isothermal)

        # hysteresis (differential)
        direction = -np.sign(current)
        coeff = np.abs(ce*current*self.gamma*Q_inv)
        rhs[ptr['hyst']] = coeff*(direction*self.M_hyst(soc) - hyst)

        # RC overpotentials (differential)
        for j, pj in enumerate(ptr['eta_j'], start=1):
            Rj = getattr(self, 'R' + str(j))(soc, T_cell)
            Cj = getattr(self, 'C' + str(j))(soc, T_cell)
            rhs[pj] = -sv[pj] / (Rj*Cj) + current / Cj

        # cell voltage (algebraic) - only if using Simulation, not Prediction
        if self._classname == 'Simulation':
            mode = userdata['mode']
            units = userdata['units']
            value = userdata['value']

            if mode == 'current' and units == 'A':
                rhs[ptr['V_cell']] = current - value(t)
            elif mode == 'current' and units == 'C':
                rhs[ptr['V_cell']] = current - self.capacity*value(t)
            elif mode == 'voltage':
                rhs[ptr['V_cell']] = voltage - value(t)
            elif mode == 'power':
                rhs[ptr['V_cell']] = power - value(t)

            # values for eventsfn
            total_time = self._t0 + t

            events = {
                'soc': soc,
                'temperature_K': T_cell,
                'current_A': current,
                'current_C': current / self.capacity,
                'voltage_V': voltage,
                'power_W': power,
                'capacity_Ah': soc*self.capacity,
                'time_s': total_time,
                'time_min': total_time / 60.,
                'time_h': total_time / 3600.,
            }

            userdata['events'] = events

        return rhs


def _yaml_reader(file: str) -> dict:
    """
    Read a yaml file.

    This yaml reader has a custom ``!eval`` constructor, which allows you to
    evaluate ``np`` expressions and ``lambda`` functions.

    Parameters
    ----------
    file : str
        An absolute or relative path to a '.yaml' file. The extension will
        be added if not included.

    Raises
    ------
    ValueError
        Invalid file. Only supports '.yaml' files.
    FileNotFoundError
        File does not exist.

    Returns
    -------
    data : dict
        Data dictionary corresponding to the input file.

    """

    _, extension = os.path.splitext(file)

    if extension == '':
        file += '.yaml'
    elif extension != '.yaml':
        raise ValueError("Invalid file. Only supports '.yaml' files.")

    here = os.path.dirname(__file__)
    templates = here + '/templates'

    if file in os.listdir(templates):
        short_warn(f"Using the default parameter file '{file}'.")
        file = templates + '/' + file

    def eval_constructor(loader, node):
        return eval(node.value)

    if not os.path.exists(file):
        raise FileNotFoundError(f"File '{file}' does not exist.")

    reader = YAML(typ='safe')

    add_constructor('!eval', eval_constructor, constructor=SafeConstructor)

    with open(file, 'r') as f:
        data = reader.load(f)

    return data


def formatwarning(message, category, filename, lineno, line=None):
    """Shortened warning format - used for parameter/pre warnings."""
    return f"\n[thevenin {category.__name__}] {message}\n\n"


def short_warn(message, category=UserWarning, stacklevel=1, source=None):
    """Print a warning with the short format from ``formatwarning``."""
    original_format = warnings.formatwarning

    warnings.formatwarning = formatwarning
    warnings.warn_explicit(message, category, filename='None', lineno=0)

    warnings.formatwarning = original_format



from __future__ import annotations
from typing import TypeVar, TYPE_CHECKING

import time
from copy import deepcopy

import numpy as np

from thevenin._basemodel import BaseModel

if TYPE_CHECKING:  # pragma: no cover
    from ._experiment import Experiment
    from ._solutions import BaseSolution, StepSolution, CycleSolution

    Solution = TypeVar('Solution', bound='BaseSolution')


class Simulation(BaseModel):
    """
    Simulation model wrapper.

    This class is primarily intended for full timeseries simulations. Use the
    :class:`~thevenin.Experiment` class to provide the details of an experiment.
    Note that this version of the model interface manages its own internal
    state. The user has limited ability to directly control the state. At the
    beginning of all simulations, the model is assumed in a fully rested state
    at the user-supplied state-of-charge ``soc0``. Through the pre-processor
    method ``pre()`` you can manually force the state to start at a value given
    by a previous solution, but you cannot individually overwrite and set any
    internal state variables. If you are interested in having more control,
    see the :class:`~thevenin.Prediction` class instead, which is intended more
    for step-by-step predictions used in prediction-correction algorithms like
    Kalman filters.

    """

    def pre(self, initial_state: bool | Solution = True) -> None:
        """
        Pre-process and prepare the model for running experiments.

        This method builds solution pointers, registers algebraic variable
        indices, stores the mass matrix, and initializes the hidden state.

        Parameters
        ----------
        initial_state : bool | Solution
            Control how the model state is initialized. If True (default), the
            state is set to a rested condition at 'soc0'. If False, the state
            is left alone and only internal checks are run. Given a Solution
            instance, the state is set to the final state of the solution. See
            the notes for more information.

        Returns
        -------
        None.

        Notes
        -----
        This method runs during the class initialization. It generally does not
        have to be run again unless you want to reset the internal hidden state.
        However, there is limited control over how users can set the state. It
        can either be set to a rested condition based on 'soc0', or it can be
        initialized based on a ``Solution`` instance.

        When initializing based on a Solution instance, the solution must be
        the same size as the current model. In other words, a 1RC-pair model
        cannot be initialized given a solution from a 2RC-pair circuit.

        """

        from ._solutions import BaseSolution

        self._check_RC_pairs()  # inherited from BaseModel

        ptr = {}
        ptr['soc'] = 0
        ptr['T_cell'] = 1
        ptr['hyst'] = 2
        ptr['eta_j'] = np.arange(3, 3 + self.num_RC_pairs)
        ptr['V_cell'] = self.num_RC_pairs + 3
        ptr['size'] = self.num_RC_pairs + 4

        algidx = [ptr['V_cell']]

        mass_matrix = np.ones(ptr['size'])
        mass_matrix[ptr['V_cell']] = 0.

        sv0 = np.zeros(ptr['size'])
        sv0[ptr['soc']] = self.soc0
        sv0[ptr['T_cell']] = self.T_inf / self._T_ref
        sv0[ptr['hyst']] = 0.
        sv0[ptr['eta_j']] = 0.
        sv0[ptr['V_cell']] = self.ocv(self.soc0)

        svdot0 = np.zeros_like(sv0)

        self._ptr = ptr
        self._algidx = algidx
        self._mass_matrix = mass_matrix

        self._t0 = 0.
        if isinstance(initial_state, BaseSolution):
            soln = deepcopy(initial_state)
            if soln.y[-1].size != sv0.size:
                raise ValueError("Cannot initialize state based on Solution"
                                 " object given in 'initial_state'. The model"
                                 " and solution have incompatible sizes.")

            self._sv0 = soln.y[-1].copy()
            self._svdot0 = soln.yp[-1].copy()

        elif initial_state:
            self._sv0 = sv0
            self._svdot0 = svdot0

    def run_step(self, exp: Experiment, stepidx: int) -> StepSolution:
        """
        Run a single experimental step.

        Parameters
        ----------
        expr : Experiment
            An experiment instance.
        stepidx : int
            Step index to run. The first step has index 0.

        Returns
        -------
        :class:`~thevenin.StepSolution`
            Solution to the experimental step.

        Warning
        -------
        The model's internal state is changed at the end of each experimental
        step. Consequently, you should not run steps out of order. You should
        always start with ``stepidx = 0`` and then progress to the subsequent
        steps afterward. If at any time you want to reset the internal hidden
        state back to a rested condition, use ``pre()``.

        See also
        --------
        Experiment : Build an experiment.
        StepSolution : Wrapper for a single-step solution.

        Notes
        -----
        Using ``run()`` loops through all steps in an experiment and stitches
        their solutions together. Most of the time, this is more convenient.
        However, running step-by-step provides maximum control to fine tune
        solver options. It also allows for complex analyses and/or control
        decisions to be performed between experimental steps.

        """

        from .solvers import IDASolver
        from ._solutions import StepSolution

        step = exp.steps[stepidx].copy()
        kwargs = exp._kwargs[stepidx].copy()

        if not callable(step['value']):
            value = step['value']
            step['value'] = lambda t: value

        kwargs['userdata'] = step
        kwargs['calc_initcond'] = 'yp0'
        kwargs['algebraic_idx'] = self._algidx

        if step['limits'] is not None:
            _setup_eventsfn(step['limits'], kwargs)

        solver = IDASolver(self._resfn, **kwargs)

        start = time.time()
        ida_soln = solver.solve(step['tspan'], self._sv0, self._svdot0)
        timer = time.time() - start

        soln = StepSolution(self, ida_soln, timer)

        self._t0 = soln.t[-1]
        self._sv0 = soln.y[-1].copy()
        self._svdot0 = soln.yp[-1].copy()

        return soln

    def run(self, exp: Experiment, reset_state: bool = True,
            t_shift: float = 1e-3) -> CycleSolution:
        """
        Run a full experiment.

        Parameters
        ----------
        expr : Experiment
            An experiment instance.
        reset_state : bool
            If True (default), the internal state of the model will be reset
            back to a rested condition at 'soc0' at the end of all steps. When
            False, the state does not reset. Instead it will update to match
            the final state of the last experimental step.
        t_shift : float
            Time (in seconds) to shift step solutions by when stitching them
            together. If zero the end time of each step overlaps the starting
            time of its following step. The default is 1e-3.

        Returns
        -------
        :class:`~thevenin.CycleSolution`
            A stitched solution with all experimental steps.

        Warning
        -------
        The default behavior resets the model's internal state back to a rested
        condition at 'soc0' by calling the ``pre()`` method at the end of all
        steps. This means that if you run a second experiment afterward, it
        will not start where the previous one left off. Instead, it will start
        from the original rested condition that the model initialized with. You
        can bypass this by using ``reset_state=False``, which keeps the state
        at the end of the final experimental step.

        See also
        --------
        Experiment : Build an experiment.
        CycleSolution : Wrapper for an all-steps solution.

        """

        from ._solutions import CycleSolution

        solns = []
        for i in range(exp.num_steps):
            solns.append(self.run_step(exp, i))

        soln = CycleSolution(*solns, t_shift=t_shift)

        self._t0 = 0.
        if reset_state:
            self.pre()

        return soln

    def _resfn(self, t: float, sv: np.ndarray, svdot: np.ndarray,
               res: np.ndarray, userdata: dict) -> None:
        """
        Solver-structured residuals.

        The IDASolver requires a residuals function in this exact form.
        Rather than outputting the residuals, the function returns None,
        but fills the 'res' input array with the DAE residuals.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        svdot : 1D np.array
            State variable time derivatives at time t.
        res : 1D np.array
            DAE residuals, res = M*yp - rhs(t, y).
        userdata : dict
            Dictionary detailing an experimental step.

        Returns
        -------
        None.

        """

        res[:] = self._mass_matrix*svdot - self._rhsfn(t, sv, userdata)


class _EventsFunction:
    """Event function callables."""

    def __init__(self, limits: tuple[str, float]) -> None:
        """
        This class is a generalized event function callable.

        Parameters
        ----------
        limits : tuple[str, float]
            A tuple of event function criteria arranged as ordered pairs of
            limit names and values, e.g., ('time_h', 10., 'voltage_V', 4.2).

        """

        self.keys = limits[0::2]
        self.values = limits[1::2]
        self.size = len(self.keys)

    def __call__(self, t: float, sv: np.ndarray, svdot: np.ndarray,
                 events: np.ndarray, userdata: dict) -> None:
        """
        Solver-structured event function.

        The IDASolver requires a event function in this exact form. Rather
        than outputting the events array, the function returns None, but
        fills the 'events' input array with event functions. If any 'events'
        index equals zero, the solver will exit prior to 'tspan[-1]'.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        svdot : 1D np.array
            State variable time derivatives at time t.
        events : 1D np.array
            An array of event functions. During integration, the solver will
            exit prior to 'tspan[-1]' if any 'events' index equals zero.
        userdata : dict
            Dictionary detailing an experimental step, with the 'events' key
            added and filled within the `rhs_funcs()' method.

        Returns
        -------
        None.

        """

        for i, (key, value) in enumerate(zip(self.keys, self.values)):
            events[i] = userdata['events'][key] - value


def _setup_eventsfn(limits: tuple[str, float], kwargs: dict) -> None:
    """
    Set up a event function for the IDASolver.

    The IDASolver requires two keyword arguments to be set when using event
    functions. The first is 'eventsfn' which requires a Callable. The second
    is 'num_events' with allocates memory to an array that stores the event
    function values.

    Parameters
    ----------
    limits : tuple[str, float]
        A tuple of event function criteria arranged as ordered pairs of limit
        names and values, e.g., ('time_h', 10., 'voltage_V', 4.2).
    kwargs : dict
        The IDASolver keyword argumnents dictionary. Both the 'eventsfn' and
        'num_events' keyword arguments must be added to 'kwargs'.

    Returns
    -------
    None.

    """

    eventsfn = _EventsFunction(limits)

    kwargs['eventsfn'] = eventsfn
    kwargs['num_events'] = eventsfn.size



from __future__ import annotations
from typing import Callable

import numpy as np

from thevenin._basemodel import BaseModel


class TransientState:
    """Transient state for predictions."""

    def __init__(self, soc: float, T_cell: float, hyst: float,
                 eta_j: np.ndarray | None) -> None:
        """
        This class allows the user to manage the state when working with the
        :class:`~thevenin.Prediction` class. The user only has control over
        independent state variables (i.e., soc, T_cell, hyst, eta_j).

        The read-only ``voltage`` property will return None if the state was
        user defined. If instead the state was returned by the ``Prediction``
        class, the value will be the predicted voltage after a given step.

        Parameters
        ----------
        soc : float
            State of charge [-].
        T_cell : float
            Temperature of the cell [K].
        hyst : float
            Hysteresis voltage [V].
        eta_j : np.ndarray | None
            RC pair overpotentials [V].

        See also
        --------
        Prediction :
            The model wrapper that interfaces with ``TransientState``.

        """

        self.soc = soc
        self.T_cell = T_cell
        self.hyst = hyst

        if eta_j is None:
            self.eta_j = np.array([])
        else:
            self.eta_j = np.asarray(eta_j)

        self._voltage = None

        self._repr_keys = [
            'soc',
            'T_cell',
            'hyst',
            'eta_j',
        ]

    def __repr__(self) -> str:  # pragma: no cover
        """
        Return a readable repr string.

        Returns
        -------
        readable : str
            A console-readable instance representation.

        """

        keys = self._repr_keys
        values = [getattr(self, k) for k in keys]

        summary = "\n    ".join([f"{k}={v}," for k, v in zip(keys, values)])
        readable = f"TransientState(\n    {summary}\n)"

        return readable

    @property
    def num_RC_pairs(self) -> int:
        """Number of RC pairs in the circuit."""
        return len(self.eta_j)

    @property
    def voltage(self) -> float | None:
        """None if user-defined state. Otherwise, predicted voltage [V]."""
        return self._voltage

    def _set_voltage(self, voltage: float) -> None:
        """A hidden method for the 'Prediction' class to set the voltage."""
        self._voltage = voltage


class Prediction(BaseModel):
    """
    Prediction model wrapper.

    This class is primarily intended for prediction-correction algorithms,
    e.g., Kalman filters. The interface is set up to let the user manage the
    internal state via the :class:`~thevenin.TransientState` class. In addition,
    the model is only designed to run current-based loads in a step-by-step
    fashion. If you are looking to simulate more complex protocols and use the
    resulting timeseries data you should use the :class:`~thevenin.Simulation`
    class instead.

    Note that this class and the :class:`~thevenin.Simulation` class share the
    same ``params`` inputs. This is for convenience so that users can easily
    switch between the two. However, the 'soc0' value has no real meaning for
    the prediction interface. Instead, the user manages the state of charge for
    each step through the ``TransientState`` interface. This means that you can
    effectively ignore the 'soc0' input when using this class.

    """

    def pre(self) -> None:
        """
        Pre-process and prepare the model to make predictions. Specifically,
        this method builds pointers so it can manage mapping the state back
        and forth between the solver-required array format and the user-managed
        ``TransientState`` class.

        Returns
        -------
        None.

        Notes
        -----
        This method runs during the class initialization. It generally does not
        have to be run again unless you want to re-run internal checks on the
        class instance.

        """

        self._check_RC_pairs()  # inherited from BaseModel

        ptr = {}
        ptr['soc'] = 0
        ptr['T_cell'] = 1
        ptr['hyst'] = 2
        ptr['eta_j'] = np.arange(3, 3 + self.num_RC_pairs)
        ptr['size'] = self.num_RC_pairs + 3

        self._ptr = ptr

        self.set_options()

    def set_options(self, **options) -> None:
        """
        Set the solver options for the underlying ODE integrator.

        Parameters
        ----------
        **kwargs : dict, optional
            CVODESolver keyword arguments that span all steps. You can re-run
            this method between prediction steps if you need different settings
            per step.

        See also
        --------
        ~thevenin.solvers.CVODESolver :
            The solver class, with documentation for most keyword arguments
            that you might want to adjust.

        """

        from .solvers import CVODESolver

        self._userdata = {}
        options = {'userdata': self._userdata, **options}
        self._solver = CVODESolver(self._svdot, **options)

    def take_step(self, state: TransientState, current: float | Callable,
                  delta_t: float) -> TransientState:
        """
        Take a step forward in time to predict the new state and voltage given
        a starting state, demand current, and time step.

        Parameters
        ----------
        state : TransientState
            Description of the starting state.
        current : float | Callable
            Demand current [A]. For a dynamic current, use a callable with a
            signature like ``def current(t: float) -> float``, where the input
            time is in seconds relative to the overall step.
        delta_t : float
            Magnitude of time step, in seconds.

        Returns
        -------
        :class:`~thevenin.TransientState`
            Predicted state at the end of the time step.

        """

        if callable(current):
            self._userdata['current'] = current
        else:
            self._userdata['current'] = lambda t: current

        sv0 = self._to_array(state)

        _ = self._solver.init_step(0., sv0)
        soln = self._solver.step(delta_t)

        # state prediction
        state = self._to_state(soln.y)

        # voltage prediction
        ocv = self.ocv(state.soc)
        R0 = self.R0(state.soc, state.T_cell)

        current = self._userdata['current'](soln.t)
        voltage = ocv + state.hyst - np.sum(state.eta_j) - current*R0

        state._set_voltage(voltage)

        return state

    def _to_state(self, array: np.ndarray) -> TransientState:
        """
        Map an array returned by the solver to a 'TransientState' instance.

        Parameters
        ----------
        array : 1D np.array
            State variables array compatible with the solver..

        Returns
        -------
        :class:`~thevenin.TransientState`
            Human-interpretable instance of the state variables array.

        """

        ptr = self._ptr.copy()
        _ = ptr.pop('size')

        state = {}
        for k, v in ptr.items():
            state[k] = array[v] * (self._T_ref if k == 'T_cell' else 1.)

        return TransientState(**state)

    def _to_array(self, state: TransientState) -> np.ndarray:
        """
        Map a 'TransientState' instance to a numpy array for the solver.

        Parameters
        ----------
        state : TransientState
            Human-interpretable instance of the state variables array.

        Returns
        -------
        array : 1D np.array
            State variables array compatible with the solver.

        Raises
        ------
        ValueError
            state.eta_j has an invalid length. The number of overpotentials
            must be consistent with the model's 'num_RC_pairs' attribute.

        """

        if state.num_RC_pairs != self.num_RC_pairs:
            raise ValueError(f"{state.eta_j=} has an invalid length since"
                             f" num_RC_pairs={self.num_RC_pairs}.")

        ptr = self._ptr.copy()
        size = ptr.pop('size')

        sv = np.zeros(size)
        for k, v in ptr.items():
            sv[v] = getattr(state, k) / (self._T_ref if k == 'T_cell' else 1.)

        return sv

    def _svdot(self, t: float, sv: np.ndarray, svdot: np.ndarray,
               userdata: dict) -> None:
        """
        Solver-structured right-hand-side.

        The CVODESolver requires a right-hand-side function in this form.
        Rather than outputting the derivatives, the function returns None,
        but fills the 'svdot' input array with the ODE expressions.

        Parameters
        ----------
        t : float
            Value of time [s].
        sv : 1D np.array
            State variables at time t.
        svdot : 1D np.array
            State variable time derivatives from ODEs, svdot = rhs(t, sv).
        userdata : dict
            Dictionary detailing an experimental step.

        Returns
        -------
        None.

        """

        svdot[:] = self._rhsfn(t, sv, userdata)
